// handle all output
#include"IO.h"
#define PI 3.1415926536
#include<cmath>
using namespace std;

extern int nmobile, numclusters, numimmclusters, ninstrings, numstrings;
extern vector <int> mobileparticleids, sizecounts, immpsizecounts, stringlengthcounts;
extern vector <double> particledisps, RgSqVsSize, immpRgSqVsSize, StringRgSqVsLength;

// Add if statements ~ min_flag, vhcflag, gofrflag, etc.
void IO::openfiles() {
    char tname[80];
    sprintf(tname,"thermodata.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
    thermout.open(tname);
    char xname[80];
    sprintf(xname,"coords.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
    xout.open(xname);
    char dname[80];
    sprintf(dname,"msdngpfov.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
    dynout.open(dname);
    char mname[80];
    if (minflag) {
        sprintf(mname,"mindata.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
        minout.open(mname);
    }
    char vname[80];
    if (vhcflag) {
        sprintf(vname,"vanHoveCorr.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
        vhcout.open(vname);
    }
    char gname[80];
    if (gofrflag) {
        sprintf(gname,"gofr.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
        gofrout.open(gname);
    }
    char mpname[100];
    if (mobileflag) {
        sprintf(mpname,"mobileparticlecoords.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
        mpout.open(mpname);
        sprintf(mpname,"mobileclustersizedistribs.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
        mpSout.open(mpname);
        sprintf(mpname,"mobileclusterRgVsSize.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
        mpRGout.open(mpname);
        sprintf(mpname,"immobileclustersizedistribs.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
        immpSout.open(mpname);
        sprintf(mpname,"immobileclusterRgVsSize.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
        immpRGout.open(mpname);
        sprintf(mpname,"stringlengthdistribs.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
        strSout.open(mpname);
        sprintf(mpname,"stringRgVsLength.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
        strRGout.open(mpname);
    }
    return;
}


// Close all files
void IO::closefiles() {
    thermout.close();
    xout.close();
    dynout.close();
    if (minflag) minout.close();
    if (vhcflag) vhcout.close();
    if (gofrflag) gofrout.close();
    if (mobileflag) {
        mpout.close();   mpSout.close();     mpRGout.close();      immpSout.close();     immpRGout.close();
        strSout.close();  strRGout.close();
    }
    return;
}


// Write sigma, coordinates, and image flags to config-dump file
void IO::writecoords(int step) {
    int i, j;
    xout.precision(10);
    xout << step << " " << lo << " " << hi << endl;
    for (i = 0; i < N; i++) {
        xout << i << " " << sigma[i];
        for (j = 0; j < d; j++) xout << " " << RVF[3*d*i+j];
        for (j = 0; j < d; j++) xout << " " << image[d*i+j];
        xout << endl;
    }
    return;
}


// Write sigma, coordinates, and displacements of mobile particles to config-dump file
void IO::writemobilecoords(int step) {
    int i, j, k;
    mpout.precision(10);
    mpout << step << " " << lo << " " << hi << endl;
    for (i = 0; i < nmobile; i++) {
        k = mobileparticleids[i];
        mpout << k << " " << sigma[k];
        for (j = 0; j < d; j++) mpout << " " << RVF[3*d*k+j];
        mpout << " " << particledisps[k] << endl;
    }
    return;
}



// Write mass, sigma, epsilon, coords, velocities, image flags to restart file
void IO::writerestart(int step) {
    int i, j;
    char rname[80];
    sprintf(rname,"restart.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f.step%d",d,N,phi,Ttarg,attfrac,step);
    resout.open(rname);
    resout.precision(10);
    resout << step << " " << lo << " " << hi << endl;
    for (i = 0; i < N; i++) {
        resout << i << " " << mass[i] << " " << sigma[i] << " " << epsilon[i];
        for (j = 0; j < 2*d; j++) resout << " " << RVF[3*d*i+j];
        for (j = 0; j < d; j++) resout << " " << image[d*i+j];
        resout << endl;
    }
    resout.close();
    return;
}



// Write initial coords, image flags
void IO::writeRinit(int step) {
    int i, j;
    char inifname[80];
    sprintf(inifname,"Rinit.d%d.N%d.phi%.4f.T%.3f.attfrac%.3f",d,N,phi,Ttarg,attfrac);
    initout.open(inifname);
    initout.precision(10);
    initout << step << " " << lo << " " << hi << endl;
    for (i = 0; i < N; i++) {
        initout << i;
        for (j = 0; j < d; j++) initout << " " << RVF[3*d*i+j];
        for (j = 0; j < d; j++) initout << " " << image[d*i+j];
        initout << endl;
    }
    initout.close();
    return;
}


// Output thermodynamic data
void IO::outputthermo(int step) {
    if (step == startstep) thermout << "Step elapstime T rho phi P Z Epair msd ngp f_ov S(q,t) avgmobileclustersize avgimmobileclustersize stringfrac avgstringsize" << endl;
    thermout << step << " " << elapstime << " " << T << " " << rho << " " << phi  << " " << P << " " << Z << " " << EP << " " << msd << " " << ngp << " " << f_ov <<  " " << SISF << " " << avgmobileclustersize << " " << avgimmobileclustersize << " " << stringfrac << " " << avgstringsize << endl;
    return;
}


// Output energy-minimization data
void IO::outputmin(int step, int iteration, double fnorm, double fracDE) {
    if (iteration == 0) minout << "Step Iteration T Epair Pressure Z Z_noratt f_norm/(d*N) EP_del/EP Rattlers Floaters namsd ngp f_ov time" << endl;
    if (step >= 0 && iteration >= 0) {
        minout << step << " " << iteration << " " << T << " " << EP << " " << P << " " << Z << " " << Znoratt << " ";
        minout << fnorm/(d*N) << " " << fracDE << " " << rattlers << " " << floaters << " " << msd << " " << ngp << " " << f_ov << " " << min_actual_time << endl;
    }
    return;
}


// Output van Hove correlation function
void IO::outputVHC(double *vhc, double *vhcsmall, double *vhclarge, double *vhcmobile, double *vhcimmobile, int step) {
    int i;
    if (step == startstep) vhcout << "r G_s(all) G_s (large) G_s (small) G_s (mobile) G_s(immobile)" << endl;
    vhcout << step << endl;
    for (i = 0; i < vhcap; i++) vhcout << (i+.5)*.01 << " " << vhc[i] << " " << vhclarge[i] << " " << vhcsmall[i] << " " << vhcmobile[i] << " " << vhcimmobile[i] << endl;
    return;
}



// Output dynamics data for the large and small particles
void IO::outputbidispdyn(int step) {
    if (step == startstep) dynout << "Step elapstime MSD (all) NGP (all) f_ov (all) SISF (mean) MSD (large) NGP (large) f_ov (large) SISF (large) MSD (small) NGP (small) f_ov (small) SISF (small) MSD (mobile) NGP (mobile) f_ov (mobile) SISF (mobile) MSD (immobile) NGP (immobile) f_ov (immobile) SISF (immobile)" << endl;
    dynout << step << " " << elapstime << " " << msd << " " << ngp << " " << f_ov << " " << SISF << " " << msd1 << " " << ngp1 << " " << fov1 << " " << sisf1 << " " << msd2 << " " << ngp2 << " " << fov2 << " " << sisf2 << " " << msdm << " " << ngpm << " " << fovm << " " << sisfm << " " << msdimm << " " << ngpimm << " " << fovimm << " " << sisfimm << endl;
    return;
}



// Output the pair correlation function
void IO::writegofr(int *nnforg, int *nnforgLL, int *nnforgLS, int *nnforgSS, int *nnforgMOB, int *nnforgIMM, int step) {
// prefactors for spherical-shell areas
    double as[11] = {0, 2, 2*PI, 4*PI, 2*PI*PI, 8*PI*PI/3, pow(PI,3), 16*pow(PI,3)/15, pow(PI,4)/3, 32*pow(PI,4)/105, pow(PI,5)/12};
    double rho = N/vol;
    double rhoL = Nlarge/vol;    double rhoS = Nsmall/vol;
    double rhoMOB = nmobile/vol;    double rhoIMM = nmobile/vol;
    double grdenom, grdenomLL, grdenomLS, grdenomSS, grdenomMOB, grdenomIMM;
    int i;
    gofrout << step << endl;
    for (i = 0; i < grcap; i++) {
        grdenom = as[d]*rho*pow((i+.5)*.01,d-1)*.01*N;
        grdenomLL = as[d]*rhoL*pow((i+.5)*.01,d-1)*.01*Nlarge;
        grdenomLS = as[d]*(2*rhoL*rhoS/rho)*pow((i+.5)*.01,d-1)*.01*N;
        grdenomSS = as[d]*rhoS*pow((i+.5)*.01,d-1)*.01*Nsmall;
        grdenomMOB = as[d]*rhoMOB*pow((i+.5)*.01,d-1)*.01*nmobile;
        grdenomIMM = as[d]*rhoIMM*pow((i+.5)*.01,d-1)*.01*nmobile;
        gofrout << (i+.5)*.01 << " " << 1.0*nnforg[i]/grdenom << " " << 1.0*nnforgLL[i]/grdenomLL << " " << 1.0*nnforgLS[i]/grdenomLS << " " << 1.0*nnforgSS[i]/grdenomSS << " " << 1.0*nnforgMOB[i]/grdenomMOB  << " " << 1.0*nnforgIMM[i]/grdenomIMM << endl;
    }
    return;
}



// Output mobile-particle cluster size- and Rg-distribs
void IO::writemobilestats() {
    int i;
    double largeclusterfrac = 1;
    for (i = 1; i < 101; i++) largeclusterfrac -= (1.0*sizecounts[i])/numclusters;
    mpSout << step << " " << elapstime << " ";
    for (i = 1; i < 101; i++) mpSout << i << " " << (1.0*sizecounts[i])/numclusters << " ";
    mpSout << ">100: " << largeclusterfrac << " ";
    int biggest = 0;
    for (i = 1; i < nmobile; i++) if (sizecounts[i] > 0) biggest = i;
    mpSout << "largest: " << biggest << endl;
    
    mpRGout << step << " " << elapstime << " ";
    for (i = 2; i < 102; i++) {
        if (sizecounts[i]) mpRGout << i-1 << " " << sqrt(RgSqVsSize[i]) << " ";
        else mpRGout << i-1 << " 0 ";
    }
    mpRGout << endl;
    
    return;
}



// Output immobile-particle cluster size- and Rg-distribs
void IO::writeimmobilestats() {
    int i;
    double largeclusterfrac = 1;
    for (i = 1; i < 101; i++) largeclusterfrac -= (1.0*immpsizecounts[i])/numimmclusters;
    immpSout << step << " " << elapstime << " ";
    for (i = 1; i < 101; i++) immpSout << i << " " << (1.0*immpsizecounts[i])/numimmclusters << " ";
    immpSout << ">100: " << largeclusterfrac << " ";
    int biggest = 0;
    for (i = 1; i < nmobile; i++) if (immpsizecounts[i] > 0) biggest = i;
    immpSout << "largest: " << biggest << endl;
    
    immpRGout << step << " " << elapstime << " ";
    for (i = 2; i < 102; i++) {
        if (immpsizecounts[i]) immpRGout << i-1 << " " << sqrt(immpRgSqVsSize[i]) << " ";
        else immpRGout << i-1 << " 0 ";
    }
    immpRGout << endl;
    
    return;
}



// Output string length- and Rg-distribs
void IO::writestringstats() {
    int i;
    double largestringfrac = 1;
    for (i = 1; i < 101; i++) largestringfrac -= (1.0*stringlengthcounts[i])/numstrings;
    strSout << step << " " << elapstime << " ";
    for (i = 1; i < 101; i++) strSout << i << " " << (1.0*stringlengthcounts[i])/numstrings << " ";
    strSout << ">100: " << largestringfrac << " ";
    int longest = 0;
    for (i = 0; i < ninstrings; i++) if (stringlengthcounts[i] > 0) longest = i;
    strSout << "longest: " << longest << endl;
    
    strRGout << step << " " << elapstime << " ";
    for (i = 2; i < 101; i++) {
        if (stringlengthcounts[i]) strRGout << i-1 << " " << sqrt(StringRgSqVsLength[i]) << " ";
        else strRGout << i-1 << " 0 ";
    }
    strRGout << endl;

    return;
}
