// SWAP Monte Carlo a la Ninarello et al Physical Review X (2017)
// Routine makes maxatt = attfrac*N swap attempts every time swapmove() is called
#include "SWAP.h"
#include"cellmanagement.h"
#include"utility.h"
#include"forcelaw.h"
#include"mm_malloc.h"
extern cell supercell;
extern forcelaw forces;
extern utility utils;

int maxatt;
int *ineighinds, *jneighinds;
short int *inneigh, *jnneigh;
int *ilist, *jlist, *icell, *jcell;
inline int mypowi(int a, int b);


// Setup: allocate the necessary arrays
void swapMC::setupswap() {
    srand(time(NULL));
    maxatt = static_cast<int>(attfrac*N);
    ineighinds = (int*) _mm_malloc(maxatt*maxneigh*sizeof(int),32);
    inneigh = (short int*) _mm_malloc(maxatt*sizeof(short int),16);
    jneighinds = (int*) _mm_malloc(maxatt*maxneigh*sizeof(int),32);
    jnneigh = (short int*) _mm_malloc(maxatt*sizeof(short int),16);
    ilist = (int*) _mm_malloc(maxatt*sizeof(int),32);
    jlist = (int*) _mm_malloc(maxatt*sizeof(int),32);
    icell = (int*) _mm_malloc(maxatt*sizeof(int),32);
    jcell = (int*) _mm_malloc(maxatt*sizeof(int),32);
    return;
}

// Deallocate these arrays
void swapMC::stopswap() {
    _mm_free(ineighinds);    _mm_free(inneigh);
    _mm_free(jneighinds);    _mm_free(jnneigh);
    _mm_free(ilist);        _mm_free(jlist);
    _mm_free(icell);        _mm_free(jcell);
    return;
}


// Attempt swaps
void swapMC::swapmove() {
    int f, g, k, u, v;
    int thisi, thisj;
    double ep, facij, delc0[d], delc[d], del, delsq;
    double ran01;  // random number 0-1 used for Metropolis acceptance/rejection
    double sigmaij, sigmaijsq, epsilonij; // for polydisp force-mixing
    double Eicurr, Ejcurr, Eiposs, Ejposs, deltaE;  // current and proposed pair energies associated with atoms thisi, thisj
    double sigmaitmp, epsilonitmp, massitmp, sigmajtmp, epsilonjtmp, massjtmp;  // for switching particles masses, "charges", diameters if swap move is accepted
    double RVFitmp[d], RVFjtmp[d];
// Generate the indices of atoms to attempt to swap and build their NLs
    if (ncell > 3) buildNLs();      else buildNsq();
    
// Attempt maxatt swaps
    for (f = 0; f < maxatt; f++) {
        Eicurr = 0;     Ejcurr = 0;     Eiposs = 0;     Ejposs = 0;
        thisi = ilist[f];       thisj = jlist[f];

// Don't attempt swaps if both thisi and thisj have the same diameter
        if (sigma[thisi]==sigma[thisj]) continue;
        
// Calculate current pair energy associated with atom thisi and the pair energy if it were swapped with atom thisj
// Loop over particle i's neighbors
        for (v = 0; v < d; v++) delc0[v] = -RVF[3*d*thisi+v];
        for (g = 0; g < inneigh[f]; g++) {
// Calculate current energy
// Apply mixing rules for particle diameters, force/energy prefactors
            u = ineighinds[f*maxneigh + g];
            sigmaij = forces.getsigmaij2(thisi,u);
            epsilonij = forces.getepsilonij2(thisi,u);
            sigmaijsq = sigmaij*sigmaij;
            delsq = 0;
            for (k = 0; k < d; k++) {
                delc[k] = delc0[k] + RVF[3*d*u+k];
                if (delc[k] < -D/2.0) delc[k] += D;    else if (delc[k] > D/2.0) delc[k] -= D;
                delsq += delc[k]*delc[k];
            }
            if (delsq < cutsq*sigmaijsq) {
                ep = forces.getepair(delsq,sigmaij,sigmaijsq,epsilonij);
                Eicurr += ep;
            }

// Calculate energy if swap is made
            sigmaij = forces.getsigmaij2(thisj,u);
            epsilonij = forces.getepsilonij2(thisj,u);
            sigmaijsq = sigmaij*sigmaij;
            if (delsq < cutsq*sigmaijsq) {
                ep = forces.getepair(delsq,sigmaij,sigmaijsq,epsilonij);
                Eiposs += ep;
            }
        }
        
// Calculate current pair energy associated with atom thisj and the pair energy if it were swapped with atom thisi
// Loop over particle i's neighbors
        for (v = 0; v < d; v++) delc0[v] = -RVF[3*d*thisj+v];
        for (g = 0; g < jnneigh[f]; g++) {
// Calculate current energy
            u = jneighinds[f*maxneigh + g];
            sigmaij = forces.getsigmaij2(thisj,u);
            epsilonij = forces.getepsilonij2(thisj,u);
            sigmaijsq = sigmaij*sigmaij;
            delsq = 0;
            for (k = 0; k < d; k++) {
                delc[k] = delc0[k] + RVF[3*d*u+k];
                if (delc[k] < -D/2.0) delc[k] += D;    else if (delc[k] > D/2.0) delc[k] -= D;
                delsq += delc[k]*delc[k];
            }
            if (delsq < cutsq*sigmaijsq) {
                ep = forces.getepair(delsq,sigmaij,sigmaijsq,epsilonij);
                Ejcurr += ep;
            }
// Calculate energy if swap is made
            sigmaij = forces.getsigmaij2(thisi,u);
            epsilonij = forces.getepsilonij2(thisi,u);
            sigmaijsq = sigmaij*sigmaij;
            if (delsq < cutsq*sigmaijsq) {
                ep = forces.getepair(delsq,sigmaij,sigmaijsq,epsilonij);
                Ejposs += ep;
            }
        }
        
// Here is the actual Metropolis MC
       deltaE = Eiposs + Ejposs - Eicurr - Ejcurr;
       ran01 = (1.0*rand())/RAND_MAX;
// Perform the swap if appropriate
        if (ran01 < exp(-deltaE/Ttarg)) {
           sigmaitmp = sigma[thisj];    epsilonitmp = epsilon[thisj];       massitmp = mass[thisj];
           sigmajtmp = sigma[thisi];    epsilonjtmp = epsilon[thisi];       massjtmp = mass[thisi];
           sigma[thisi] = sigmaitmp;    epsilon[thisi] = epsilonitmp;       mass[thisi] = massitmp;
           sigma[thisj] = sigmajtmp;    epsilon[thisj] = epsilonjtmp;       mass[thisj] = massjtmp;
           sqrtepsilon[thisi] = sqrt(epsilon[thisi]);
           sqrtepsilon[thisj] = sqrt(epsilon[thisj]);
// Rescale velocities to conserve kinetic energy
           for (k = 0; k < d; k++) {
               RVF[3*d*thisi + d + k] *= sqrt(massjtmp/massitmp);
               RVF[3*d*thisj + d + k] *= sqrt(massitmp/massjtmp);
           }
           swapsucc++;
       }
       swapatt++;
   } // End of the f-loop
// If necessary, reorder atom indices to maintain the proper distinction between large, small particles
    if (swapsucc > 0) utils.reorderatoms();
    
    return;
}



// Build the full (non-N3) NLs for the atoms we will attempt to swap
// This routine is modeled on cellmanagement.cpp's buildneighborlist()
// But the NL cutoff always uses the MAXIMUM value, to avoid the issue of neglecting to count the interactions introduced by swapping in a larger particle
void swapMC::buildNLs() {
    int att, f, l, p, q, r, v;
    int thisicell, thisjcell;
    int powofnc[d];
    for (f = 0; f < d; f++) powofnc[f] = mypowi(ncell,f); // need have this here rather than in setupswap() because ncell isn't always constant
    int neighindex[numneighcells]; // same as in cellmanagement()
    int neighpos[numneighcells*d];  // same as in cellmanagement()
    int toi[d], toj[d];  // like to[d] in buildneighborlist()
    double delc[d], delc0[d], delsq;  // same as in buildneighborlist()
    double sigmaqr, rangeqr;
    if (model == 1 || model == 2) sigmaqr = ALU[2];
    else if (model == 3) sigmaqr = sAA;
    rangeqr = cut*sigmaqr + skin;
    
// Generate the i-j swap attempt pairs: select particles i, j randomly
    for (att = 0; att < maxatt; att++) {
        ilist[att] = static_cast<int>(((1.0*rand())/RAND_MAX)*N);
        jlist[att] = static_cast<int>(((1.0*rand())/RAND_MAX)*N);
    }
    
// Determine which cells the randomly selected atoms belong to
    #pragma omp parallel for private(thisicell,thisjcell,f,toi,toj)
    for (att = 0; att < maxatt; att++) {
        thisicell = 0;      thisjcell = 0;
        for (f = 0; f < d; f++) {
            toi[f] = floor(((RVF[3*d*ilist[att]+f] - lo)/(D/ncell)));
            toj[f] = floor(((RVF[3*d*jlist[att]+f] - lo)/(D/ncell)));
            thisicell += toi[f]*powofnc[d-1-f];
            thisjcell += toj[f]*powofnc[d-1-f];
        }
        icell[att] = thisicell;     jcell[att] = thisjcell;
    }
    
// Reset the i-atoms' and j-atoms' NLs
    #pragma omp parallel for private(l)
    for (f = 0; f < maxatt; f++) {
        inneigh[f] = 0;     jnneigh[f] = 0;
        for (l = 0; l < maxneigh; l++) {
            ineighinds[f*maxneigh + l] = -1;        jneighinds[f*maxneigh + l] = -1;
        }
    }
         
// Build NLs for the i-atoms
    #pragma omp parallel for schedule(guided) private(l,q,r,v,delc,delc0,delsq,neighindex,neighpos)
    for (f = 0; f < maxatt; f++) {
        for (l = 0; l < numneighcells; l++) {
            neighindex[l] = 0;
            for (v = 0; v < d; v++) {
                neighpos[l*d+v] = cellpos[icell[f]][v] + relneighshift[l][v];
                if (neighpos[l*d+v] == -1) neighpos[l*d+v] = ncell - 1;
                else if (neighpos[l*d+v] == ncell) neighpos[l*d+v] = 0;
// Additional cases for cells of size (r_c + skin)/2-typdist and nbs == 5
                if (nbs == 5 && neighpos[l*d+v] == -2) neighpos[l*d+v] = ncell - 2;
                else if (nbs == 5 & neighpos[l*d+v] == ncell+1) neighpos[l*d+v] = 1;
            }
            for (v = 0; v < d; v++) neighindex[l] += neighpos[l*d+v]*powofnc[d-1-v];
        }
// Loop over all subcells neighboring that belonging to atom q
// Below is similar to method used in buildneighborlist()
        q = ilist[f];
        for (v = 0; v < d; v++) delc0[v] = -RVF[3*d*q+v];
        for (l = 0; l < numneighcells; l++) for (p = 0; p < supercell.boxpops[neighindex[l]]; p++) {
            r = supercell.subcell[neighindex[l]].atomlist[p];
            if (r != q) {
                delsq = 0;
                for (v = 0; v < d; v++) {
                    delc[v] = delc0[v] + RVF[3*d*r+v];
                    if (delc[v] < -D/2.0) delc[v] += D;    else if (delc[v] > D/2.0) delc[v] -= D;
                    delsq += delc[v]*delc[v];
                    if (d > 5) if (delsq > rangeqr*rangeqr) goto nextineigh;
                }
                if (d > 5 || delsq < rangeqr*rangeqr) {
                    ineighinds[f*maxneigh + inneigh[f]] = r;
                    inneigh[f]++;
                }
            } // end of the r != q conditional
            nextineigh: ;
        } // end of the lp-loop
    }
    
// Build NLs for the j-atoms
    #pragma omp parallel for schedule(guided) private(l,q,r,v,delc,delc0,delsq,neighindex,neighpos)
    for (f = 0; f < maxatt; f++) {
        for (l = 0; l < numneighcells; l++) {
            neighindex[l] = 0;
            for (v = 0; v < d; v++) {
                neighpos[l*d+v] = cellpos[jcell[f]][v] + relneighshift[l][v];
                if (neighpos[l*d+v] == -1) neighpos[l*d+v] = ncell - 1;
                else if (neighpos[l*d+v] == ncell) neighpos[l*d+v] = 0;
// Additional cases for cells of size (r_c + skin)/2-typdist and nbs == 5
                if (nbs == 5 && neighpos[l*d+v] == -2) neighpos[l*d+v] = ncell - 2;
                else if (nbs == 5 & neighpos[l*d+v] == ncell+1) neighpos[l*d+v] = 1;
            }
            for (v = 0; v < d; v++) neighindex[l] += neighpos[l*d+v]*powofnc[d-1-v];
        }
// Loop over all subcells neighboring that belonging to atom q
// Below is similar to method used in buildneighborlist()
        q = jlist[f];
        for (v = 0; v < d; v++) delc0[v] = -RVF[3*d*q+v];
        for (l = 0; l < numneighcells; l++) for (p = 0; p < supercell.boxpops[neighindex[l]]; p++) {
            r = supercell.subcell[neighindex[l]].atomlist[p];
            if (r != q) {
                delsq = 0;
                for (v = 0; v < d; v++) {
                    delc[v] = delc0[v] + RVF[3*d*r+v];
                    if (delc[v] < -D/2.0) delc[v] += D;    else if (delc[v] > D/2.0) delc[v] -= D;
                    delsq += delc[v]*delc[v];
                    if (d > 5) if (delsq > rangeqr*rangeqr) goto nextjneigh;
                }
                if (d > 5 || delsq < rangeqr*rangeqr) {
                    jneighinds[f*maxneigh + jnneigh[f]] = r;
                    jnneigh[f]++;
                }
            } // end of the r != q conditional
        nextjneigh: ;
        } // end of the lp-loop
        
    } // Now all the necessary NLs are built

    return;
}


// raises integer (a) to an integer power (b)
inline int mypowi(int a, int b) {
    int c, i;
    c = 1;
    for (i = 0; i < b; i++) c *= a;
    return c;
}




// Build the full (non-N3) NLs for the atoms we will attempt to swap
// This routine is modeled on cellmanagement.cpp's buildNsq()
void swapMC::buildNsq() {
    int att, f, l, p, q, r, v;
    int thisicell, thisjcell;
    double delc[d], delc0[d], delsq;  // same as in buildneighborlist()
    double rangeqr;

// Generate the i-j swap attempt pairs: select particles i, j randomly
    for (att = 0; att < maxatt; att++) {
        ilist[att] = static_cast<int>(((1.0*rand())/RAND_MAX)*N);
        jlist[att] = static_cast<int>(((1.0*rand())/RAND_MAX)*N);
    }
    
// Reset the i-atoms' and j-atoms' NLs
    #pragma omp parallel for private(l)
    for (f = 0; f < maxatt; f++) {
        inneigh[f] = 0;     jnneigh[f] = 0;
        for (l = 0; l < maxneigh; l++) {
            ineighinds[f*maxneigh + l] = -1;        jneighinds[f*maxneigh + l] = -1;
        }
    }
         
// Build NLs for the i-atoms
    #pragma omp parallel for schedule(guided) private(q,r,v,rangeqr,delc,delc0,delsq)
    for (f = 0; f < maxatt; f++) {
        q = ilist[f];
        for (v = 0; v < d; v++) delc0[v] = -RVF[3*d*q+v];
        for (r = 0; r < N; r++) {
          if (r != q) {
            rangeqr = cut*(sigma[q]+sigma[r])/2 + skin;
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] = delc0[v] + RVF[3*d*r+v];
                if (delc[v] < -D/2.0) delc[v] += D;    else if (delc[v] > D/2.0) delc[v] -= D;
                delsq += delc[v]*delc[v];
                if (d > 5) if (delsq > rangeqr*rangeqr) goto nextineigh;
            }
            if (d > 5 || delsq < rangeqr*rangeqr) {
                ineighinds[f*maxneigh + inneigh[f]] = r;
                inneigh[f]++;
            }
          } // end of the r != q conditional
          nextineigh: ;
        } // end of the r-loop
    } // end of the f-loop
    
// Build NLs for the j-atoms
    #pragma omp parallel for schedule(guided) private(q,r,v,rangeqr,delc,delc0,delsq)
    for (f = 0; f < maxatt; f++) {
        q = jlist[f];
        for (v = 0; v < d; v++) delc0[v] = -RVF[3*d*q+v];
        for (r = 0; r < N; r++) {
          if (r != q) {
            rangeqr = cut*(sigma[q]+sigma[r])/2 + skin;
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] = delc0[v] + RVF[3*d*r+v];
                if (delc[v] < -D/2.0) delc[v] += D;    else if (delc[v] > D/2.0) delc[v] -= D;
                delsq += delc[v]*delc[v];
                if (d > 5) if (delsq > rangeqr*rangeqr) goto nextjneigh;
            }
            if (d > 5 || delsq < rangeqr*rangeqr) {
                jneighinds[f*maxneigh + jnneigh[f]] = r;
                jnneigh[f]++;
            }
          } // end of the r != q conditional
          nextjneigh: ;
        } // end of the r-loop
    } // end of the f-loop
        
// Now all the necessary NLs are built
    return;
}
