// This should work in arbitrary --reasonable-- d
#include"cellmanagement.h"
#include"IO.h"
#include"forcelaw.h"
vector< vector<short int> > cellpos, relneighshift;
extern int lastswapstep;
extern double cut, cutsq;
int numneighcells;
int powersofnc[10];
inline int mypowi(int a, int b);
extern IO readwrite;
extern forcelaw forces;
extern vector <bool> ismobile, isimmobile;

// Using cellpos and relneighshift to handle neigbor-mapping significantly reduces memory required to run code with minimal CPU overhead
// Elements of cellpos are (0, 1, ..., ncell - 1) along each dimension
// Elements of relneighshift are (-1, 0, 1) along each dimension
// nbs = 3 or 5 for 3^d or 5^d neighboring cells: only use nbs = 5 for long-range interactions
void cell::setupsupercell() {
    int i, f, h;
    totcells = mypowi(ncell,d);
    subcell.resize(totcells);
//    boxpops.resize(totcells);
    boxpops = (short int*) _mm_malloc(totcells*sizeof(short int),16);
    numneighcells = mypowi(nbs,d);  // For nbs = 5 this can be substantially reduced by corner-cutting
    int shift;
    if (nbs == 3) shift = 1;
    else if (nbs == 5) shift = 2;
    
// Define cellpos and relneighpos
    cellpos.resize(totcells);
    for (i = 0; i < totcells; i++) cellpos[i].resize(d);
    relneighshift.resize(numneighcells);
    for (i = 0; i < numneighcells; i++) { relneighshift[i].resize(d); }
    for (i = 0; i < totcells; i++) {
        h = i;
        for (f = 0; f < d; f++) {
            cellpos[i][d-1-f] = h%ncell;
            h /= ncell;
        }
    }
    for (i = 0; i < numneighcells; i++) {
        h = i;
        for (f = 0; f < d; f++) {
            relneighshift[i][d-1-f] = h%nbs - shift;  // relneighshift*(D/ncell) = min-image separation vector between cell centers
            h /= nbs;
        }
    }
    
// powersofnc gets used to determine neighbor l of cell i in buildneighborlist()
    for (i = 0; i < d; i++) powersofnc[i] = mypowi(ncell,i);
//    cout << "Supercell setup completed: " << totcells << " cells generated." << endl;
    return;
}



// Build or rebuild the neighborlist: works for "arbitrary" (reasonable) d
void cell::buildneighborlist() {
    int f, h, i, j, k, l, m, n, o, p, q, r, u, v;
    int to[d];    // used to place atoms in bins
//    int to2;
    double delc[d], delc0[d], delsq;
    double psum[d];  // for zeroing momentum
    double sigmaqr, rangeqr; // for polydispersity-aware neighbor-list building
    double totneigh = 0;
// Additional variables allowing parallelization of subcell-population loop
    vector <int> subcellids;
    subcellids.resize(N);
    
// Reset the number of neighbors to zero, reset neigh-lists
    # pragma omp parallel for schedule(static,nchunk) private(v)
    for (q = 0; q < N; q++) {
        numneigh[q] = 0;
        for (v = 0; v < maxneigh; v++) allneigh[q*maxneigh + v] = -1;
    }
        
// Much of the below is unnecessary if ncell <= 3, so skip it and perform an O(N^2) build
    if ((nbs == 3 && ncell < 4) || (nbs == 5 && ncell < 6)) { buildNsq();  goto zero; }

// Reset subcell & atomic-neighbor lists
    # pragma omp parallel for schedule(static,nchunk)
    for (i = 0; i < totcells; i++) {    subcell[i].atomlist.clear();     boxpops[i] = 0;  }

// Put atoms into their subcells
// Finally parallelized this 12/29/22 -- reduces NL-build time by up to ~25%!
    # pragma omp parallel for private (f, v, to, i) //reduction(+:boxpops)
    for (q = 0; q < N; q++) {
        f = 0;
        for (v = 0; v < d; v++) {
            to[v] = floor(((RVF[3*d*q+v] - lo)/(D/ncell)));
// Cleanly exit program if an atom gets lost -- only works in serial???
/*            #pragma omp critical
            if (to[v] < 0 || to[v] > ncell-1) {
                cout << "Error on step " << step << ": Atom " << u << " escaped the simulation cell, current position is";
                for (i = 0; i < d; i++) cout << " " << RVF[3*d*q+i];
                cout << endl;
                exit(1);
            } */
            f += to[v]*powersofnc[d-1-v];
        }
        subcellids[q] = f;
    }
// Second stage has to be done serially
    for (q = 0; q < N; q++) {
        subcell[subcellids[q]].atomlist.emplace_back(q);
        boxpops[subcellids[q]] += 1;
    }
    
    subcellids.clear();
    
// Build neighbor lists by looping over the cells
//    __assume_aligned(RVF,64);
#pragma omp parallel
{
    int *neighindex;    int *neighpos;
    neighindex = (int*) _mm_malloc(numneighcells*sizeof(int),32);
    neighpos = (int*) _mm_malloc(numneighcells*d*sizeof(int),32);
    #pragma omp for schedule(guided) private(l,m,o,p,q,r,v,rangeqr,sigmaqr,delc,delc0,delsq)
    for (i = 0; i < totcells; i++) if (boxpops[i]) { // loop over all cells i;
// Apply PBCs for all neighboring cells and get their indices (neighindex)
        for (l = 0; l < numneighcells; l++) {
            neighindex[l] = 0;
            for (v = 0; v < d; v++) {
                neighpos[l*d+v] = cellpos[i][v] + relneighshift[l][v];
                if (neighpos[l*d+v] == -1) neighpos[l*d+v] = ncell - 1;
                else if (neighpos[l*d+v] == ncell) neighpos[l*d+v] = 0;
// Additional conditionals are for cells of size (r_c + skin)/2-typdist and nbs == 5
                if (nbs == 5 && neighpos[l*d+v] == -2) neighpos[l*d+v] = ncell - 2;
                else if (nbs == 5 & neighpos[l*d+v] == ncell+1) neighpos[l*d+v] = 1;
            }
            for (v = 0; v < d; v++) neighindex[l] += neighpos[l*d+v]*powersofnc[d-1-v];
        }
        
        for (l = 0; l < numneighcells; l++) if (neighindex[l] >= i && boxpops[neighindex[l]]) {
// Loop over atoms in cell i
            for(o = 0; o < boxpops[i]; o++) {
                q = subcell[i].atomlist[o];   // index of the atom code is currently mapping contacts for
                for (v = 0; v < d; v++) delc0[v] = -RVF[3*d*q+v];
// Loop over cell i's neighbors
                for (p = 0; p < boxpops[neighindex[l]]; p++) {
                    r = subcell[neighindex[l]].atomlist[p];
                    if (neighindex[l] > i || (neighindex[l] == i && r > q)) {
// Apply mixing rules for particle diameters
                        sigmaqr = forces.getsigmaij(q,r);
                        rangeqr = cut*sigmaqr + skin;
                        delsq = 0;
                        for (v = 0; v < d; v++) {
                            delc[v] = delc0[v] + RVF[3*d*r+v];
                            if (delc[v] < -D/2.0) delc[v] += D;
                            else if (delc[v] > D/2.0) delc[v] -= D;
                            delsq += delc[v]*delc[v];
                            if (d > 5) if (delsq > rangeqr*rangeqr) goto nextneigh;
// Above "nextneigh" method seems to cause a slowdown for d < 5, speedup for d > 5
                        }
                        if (d > 5 || delsq < rangeqr*rangeqr) {
                            allneigh[q*maxneigh + numneigh[q]] = r;
                            numneigh[q]++;
                        }
                    } // end of the Newton's-3rd-law conditional
                    nextneigh: ;
                } // end of the p-loop
            } // end of the o-loop
        } // end of the l-loop
    } // end of the i-loop
    _mm_free(neighindex);    _mm_free(neighpos);
} // end of the parallel pragma
    
// Zero out the total system momentum and update Rlast
    zero:
    #pragma omp single
    for (i = 0; i < d; i++) psum[i] = 0;
#pragma omp parallel private(j)
{
    #pragma omp for schedule(static,nchunk) reduction(+:psum)
    for (k = 0; k < d*N; k++) psum[k%d] += mass[k/d]*RVF[(3*d)*(k/d) + d + k%d];
    #pragma omp for schedule(static,nchunk)
    for (k = 0; k < d*N; k++) RVF[(3*d)*(k/d) + d + k%d] -= psum[k%d]/(N*mass[k/d]);
    #pragma omp for schedule(static,nchunk)
    for (i = 0; i < N; i++) for (j = 0; j < d; j++) Rlast[d*i+j] = RVF[3*d*i+j];
}
    
// Issue warning if any atoms have too many neighbors
    int tmpmaxn = 0;
    if (step > 0) {
        for (i = 0; i < N; i++) if (numneigh[i] > tmpmaxn) tmpmaxn = numneigh[i];
        if (tmpmaxn >= maxneigh) cout << "Step " << step << " maxneigh = " << maxneigh << " max numneigh = " << tmpmaxn << endl;
    }
    
    return;
}


// Simpler O(N^2) NL-build for when ncell <= 3
void cell::buildNsq() {
    int q, r, v;
    double delc[d], delc0[d], delsq;
    double sigmaqr, rangeqr, rangeqrsq;
// The above variables all have the same uses as in buildneighborlist()
    #pragma omp parallel for schedule(guided) private(r,v,rangeqr,delsq,delc0,delc)
    for (q = 0; q < N-1; q++) {
        for (v = 0; v < d; v++) delc0[v] = -RVF[3*d*q+v];
        for (r = q+1; r < N; r++) {
// Apply mixing rules for particle diameters
            sigmaqr = forces.getsigmaij(q,r);
            rangeqr = cut*sigmaqr + skin;
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] = delc0[v] + RVF[3*d*r+v];
                if (delc[v] < -D/2.0) delc[v] += D;    else if (delc[v] > D/2.0) delc[v] -= D;
                delsq += delc[v]*delc[v];
                if (delsq > rangeqr*rangeqr) goto nextr;
            }
            if (delsq < rangeqr*rangeqr) {
                allneigh[q*maxneigh + numneigh[q]] = r;
                numneigh[q]++;
            }
            nextr: ;
        } // end of the r-loop
    } // end of the q-loop
    
    return;
}


// Check to see if the neighbor list needs to be rebuilt
// Rebuild NL if more than .1% of atoms have moved by at least half the skin depth
// or any atoms have moved by at least the skin depth since the previous build
bool cell::needsrebuild() {
    bool build = false;
    int i, j, k, kmax;
    double delc[d], delc0[d], delsq;
    double delsqmax = 0;
    k = 0;  // rebuild when k > .001N
    #pragma omp parallel for private(j,delc0,delc,delsq) reduction(+:k)
    for (i = 0; i < N; i++) {
            for (j = 0; j < d; j++) delc0[j] = - Rlast[d*i+j];
            for (j = 0; j < d; j++) delc[j] = delc0[j] + RVF[3*d*i+j];
            for (j = 0; j < d; j++) {
                if (delc[j] < -D/2.0) delc[j] += D;    else if (delc[j] > D/2.0) delc[j] -= D;
            }
            delsq = 0;
            for (j = 0; j < d; j++) delsq += delc[j]*delc[j];
            if (delsq > skin*skin/4) k++;
            if (delsq > skin*skin) build = true;
            if (delsq > delsqmax) delsqmax = delsq;
    }
    kmax = .001*N;
    if (k >= kmax) build = true;
    return build;
}



inline int mypowi(int a, int b) {
    int c, i;
    c = 1;
    for (i = 0; i < b; i++) c *= a;
    return c;
}



// Calculate the pair correlation function using an O(N) implementation that requires rebuilding the atom lists with the minimum # of cells necessary to calculate g(r) out to r = paircorrcap
void cell::calcgofr() {
    int f, h, i, j, k, l, m, n, o, p, q, r, u, v;
    int to[d];    // used to place atoms in bins
    double delc[d], delc0[d], delsq, del;
    int thisneighbin;
    int nnforg[grcap], nnforgLL[grcap], nnforgLS[grcap], nnforgSS[grcap], nnforgMOB[grcap], nnforgIMM[grcap];
    for (i = 0; i < grcap; i++) {
        nnforg[i] = 0;      nnforgLL[i] = 0;        nnforgLS[i] = 0;        nnforgSS[i] = 0;    nnforgMOB[i] = 0;   nnforgIMM[i] = 0;
    }
// Reset subcell & atomic-neighbor lists
    for (i = 0; i < totcells; i++) {    subcell[i].atomlist.clear();     boxpops[i] = 0;  }
// Put atoms into their subcells
    for (q = 0; q < N; q++) {
        f = 0;
        for (v = 0; v < d; v++) {
            to[v] = floor(((RVF[3*d*q+v] - lo)/(D/ncell)));
// Cleanly exit program if an atom gets lost
            if (to[v] < 0 || to[v] > ncell-1) {
                cout << "Error on step " << step << ": Atom " << u << " escaped the simulation cell, current position is";
                for (i = 0; i < d; i++) cout << " " << RVF[3*d*q+i];
                cout << endl;
                exit(1);
            }
            f += to[v]*powersofnc[d-1-v];
        }
        subcell[f].atomlist.emplace_back(q);
        boxpops[f] += 1;
    }

// Don't need to build NLs here
#pragma omp parallel
{
    int *neighindex;    int *neighpos;
    neighindex = (int*) _mm_malloc(numneighcells*sizeof(int),32);
    neighpos = (int*) _mm_malloc(numneighcells*d*sizeof(int),32);
    #pragma omp for schedule(guided) private(l,m,o,p,q,r,v,del,delc,delc0,delsq,thisneighbin) reduction(+:nnforg,nnforgLL,nnforgLS,nnforgSS)
    for (i = 0; i < totcells; i++) if (boxpops[i]) { // loop over all cells i;
// Apply PBCs for all neighboring cells and get their indices (neighindex)
// This function always uses nbs = 3
        for (l = 0; l < numneighcells; l++) {
            neighindex[l] = 0;
            for (v = 0; v < d; v++) {
                neighpos[l*d+v] = cellpos[i][v] + relneighshift[l][v];
                if (neighpos[l*d+v] == -1) neighpos[l*d+v] = ncell - 1;
                else if (neighpos[l*d+v] == ncell) neighpos[l*d+v] = 0;
            }
            for (v = 0; v < d; v++) neighindex[l] += neighpos[l*d+v]*powersofnc[d-1-v];
        }
        
        for (l = 0; l < numneighcells; l++) if (neighindex[l] >= i && boxpops[neighindex[l]]) {
// Loop over atoms in cell i
            for(o = 0; o < boxpops[i]; o++) {
                q = subcell[i].atomlist[o];   // index of the atom code is currently mapping contacts for
                for (v = 0; v < d; v++) delc0[v] = -RVF[3*d*q+v];
// Loop over cell i's neighbors
                for (p = 0; p < boxpops[neighindex[l]]; p++) {
                    r = subcell[neighindex[l]].atomlist[p];
                    if (neighindex[l] > i || (neighindex[l] == i && r > q)) {
                        delsq = 0;
                        for (v = 0; v < d; v++) {
                            delc[v] = delc0[v] + RVF[3*d*r+v];
                            if (delc[v] < -D/2.0) delc[v] += D;
                            else if (delc[v] > D/2.0) delc[v] -= D;
                            delsq += delc[v]*delc[v];
                            if (d > 5) if (delsq > paircorrcap*paircorrcap) goto nextneigh;
// Above "nextneigh" method seems to cause a slowdown for d < 5, speedup for d > 5
                        }
                        if (d > 5 || delsq < paircorrcap*paircorrcap) {
                            del = sqrt(delsq);
                            thisneighbin = floor(del/.01);
                            if (thisneighbin > grcap - 1) goto nextneigh;
                            nnforg[thisneighbin] += 2;
                            if (q < Nlarge && r < Nlarge) nnforgLL[thisneighbin] += 2;
                            else if ((q < Nlarge && r >= Nlarge) || (q >= Nlarge && r < Nlarge)) nnforgLS[thisneighbin] += 2;
                            else if (q >= Nlarge && r >= Nlarge) nnforgSS[thisneighbin] += 2;
                            if (ismobile[q] && ismobile[r]) nnforgMOB[thisneighbin] += 2;
                            if (isimmobile[q] && isimmobile[r]) nnforgIMM[thisneighbin] += 2;
                        }
                    } // end of the Newton's-3rd-law conditional
                    nextneigh: ;
                } // end of the p-loop
            } // end of the o-loop
        } // end of the l-loop
    } // end of the i-loop
    _mm_free(neighindex);    _mm_free(neighpos);
} // end of the parallel pragma
// Output g(r)
    readwrite.writegofr(nnforg,nnforgLL,nnforgLS,nnforgSS,nnforgMOB,nnforgIMM,step);
    
    return;
}
