// global.h - global variables
// Get this into a form closer to a publicly releasable version
// Document everything!
# ifndef GLOBAL_H
# define GLOBAL_H
# endif
#include<iostream>
#include<fstream>
#include<vector>
#include<ctime>
#include<cmath>
#include<cstdlib>
#include<algorithm>
#include<chrono>
#include"omp.h"
using namespace std;

// Global variables
extern double* RVF;    // current positions, velocities, forces
extern double* Rlast;  // positions at last NL build
extern double* Rinit;  // initial positions (for calculation of MSD, etc.)
extern double* mass;   // particle masses
extern double* sigma;  // particle diameters
extern double* epsilon;    // particle energy/force prefactors
extern double* sqrtepsilon;  // sqrt(epsilon)
extern int* allneigh;  // neighbor list
extern int* Z_array;   // array for counting Z
extern short int* numneigh;    // number of neighbors/atom
extern int maxneigh;   // maximum number of neighbors/atom
extern short int* image;   // current particle image flags
extern short int* initimg; // initial image flags (for calculation of MSD, etc.)
extern double ALU[3];  // polydisperse-sigma-distribution parameters
extern int d;      // spatial dimension
extern int N;      // number of particles simulated
extern int Nlarge, Nsmall;  // number of A, B particles in KA bidisperse mixture
extern int ncell;  // number of cubic subcells along each dimension
extern int ncellprev;  // previous value of ncell (needed if cell volume V varies)
extern int totcells;   // total number of subcells
extern int nbs;    // nbs = 3 or 5 controls structure of cell linking
extern double skin;    // skin depth for NL builds
extern double cut, cutsq;  // for longer-range interactions
extern int neighlistbuilds;    // number of NL builds performed
extern double avgneigh, totavgneigh;   // current and overall average neighbors/particle
extern int nthreads;   // parallel simulations run on nthreads cores
extern const int nchunk;     // Use schedule(static,nchunk) for integration of all EOM
extern double dt, dthalf;      // timestep, half-timestep for leapfrog, etc.
extern double elapstime;  // "Lennard-Jones" time since startstep or nmsdstart
extern int nsteps;     // simulation runs for nsteps timesteps
extern int nbet;       // leapfrog this many steps between NL-build-checks
extern int nthermo;    // calculate thermodynamic data every nthermo timesteps
extern int nwindow;    // average thermo data over time windows of length nwindow timesteps
extern int nvhc;       // calculate van Hove corr. function every nvhc timesteps
extern int nbaro;      // apply barostat every nbaro timesteps
extern int nmin;       // energy-minimize every nmin timesteps
extern int nswap;       // SWAP every nmin timesteps
extern int nmsdstart;  // start MSD calculation after nmsdstart steps
extern int thiswindowstart;  // timestep at start of the current "window"
extern int nrestart;   // write restart file every nrestart timesteps
extern int ndump;     // write coordinates every ndump timesteps
extern int nmobdump;   // write mobile-particle coords every nmobdump timesteps
extern int rattlers, floaters;  //counting rattler and floater particles
extern int step, startstep;   // current and starting timesteps
extern int lastswapstep;  // sometimes required for stability when SWAP employed
extern long long int swapatt, swapsucc;    // number of attempted & successful swaps
extern double attfrac;  // attempt swaps for attfrac of particles every time swapmove() is called
extern double D;   // simulation cell side length
extern double Dinit;   // initial value of D (for calculation of MSD, etc.)
extern double lo, hi;  // positions of cell's lower, upper boundaries (typically -/+ D/2)
extern double rho, phi;    // particle number density and packing fraction
extern double totpartvol, vol;     // total particle volume & cell volume
extern double E, EP, EK;  // total, pair and kinetic energies
extern double EXPA, facijtmp;  // for speeding up force & energy evaluation
extern double T, P, Z, Znoratt;    // temperature, pressure, average coordination number
extern double Ttarg, Ptarg;    // target temperature and pressure for thermostat, barostat
extern double Tdot; // for runs with temperature ramp
extern double msd, namsd, ngp, maxdisp, f_ov, SISF;  // measures of diffusion
extern double msd1, m4d1, ngp1, fov1, msd2, m4d2, ngp2, fov2;  // for bidisperse systems
extern double msdm, m4dm, ngpm, fovm, msdimm, m4dimm, ngpimm, fovimm;  // for mobile and immobile particles
extern double sisf1, sisf2, sisfm, sisfimm;  // self-intermediate scattering functions of large, small, mobile particles
extern double avgmobileclustersize, avgimmobileclustersize, stringfrac, avgstringsize;  // for mobile-particle-cluster analyses
extern double min_actual_time; //measure of actual time in FIRE and total displacement in SD
extern double paircorrcap, VHCcap; // Maximum distance out to which to calculate g(r) and G_s(r)
extern int grcap, vhcap;  // number of bins in the output g(r)
extern int *nnforg, *nnforgLL, *nnforgLS, *nnforgSS;    // binned # of neighbors vs. distance
extern bool baroflag, minflag, vhcflag, gofrflag, sofqflag, mobileflag;
extern const int model;  // 1 for GLJ, 2 for MORSE, 3 for KOBAND
extern int alpha, n;  // Morse, GLJ exponents
extern const double sAA; // sigma for the A-particles in a Kob-Andersen simulation
