// initialize.cpp -- initialize RVF, ALU, m, sigma, epsilon, etc.
#include"utility.h"
#define PI 3.1415926536
using namespace std;
extern double ALU[3];
extern double cut;
void reorderatoms(); // If necessary, reorder the atom ids


// Intialize position, force, velocity, image, neighbor arrays
// Initialize mass and sigma distributions; see RevisitingPolyDisp.nb and CheckPolyDisp.nb
// Use P(σ) = Aσ^-d form from Ninarello17: this ensures different-size particles occupy equal volume fractions
// MVR = ratio of max/minimum particle volumes: code slows down only very slowly with increasing MVR
// _mm_malloc() makes the memory occupied by the given array --aligned--
double utility::initRVFMSE(bool readresflag, const char *initresfname, const char *initRinitfname) {
    int i, j, k;
    double A, L, U;  // normalization prefactor, minimum and maximum particle diameters
    double MVR = pow(2,d/3.0);  // script-R in the paper
// This MVR gives nearly-constant PDI ~= .066 over the range 2 <= d <= 6
    L = pow(MVR,-1.0/(2*d));        U = pow(MVR,1.0/(2*d));
    A = ((d-1.0)/d)/(pow(MVR,(1.0-d)/(2*d*d))-pow(MVR,(d-1.0)/(2*d*d)));
    ALU[0] = A;     ALU[1] = L;     ALU[2] = U;
    double trialm, ranr, prob;
    double totpartvol = 0;
    int initstep;
    double initlo, inithi;
    int kissnum[11] = {0,2,6,12,24,40,72,126,240,306,500}; // kissing numbers for d-dimensional hyperspheres
    double vs[11] = {0, 1, PI/4, PI/6, PI*PI/32, PI*PI/60, pow(PI,3)/384, pow(PI,3)/840, pow(PI,4)/6144, pow(PI,4)/15120, pow(PI,5)/122880}; // volumes of d-dimensional unit hyperspheres
    
    RVF = (double*) _mm_malloc(3*d*N*sizeof(double),64);
    Rlast = (double*) _mm_malloc(d*N*sizeof(double),64);
    Rinit = (double*) _mm_malloc(d*N*sizeof(double),64);
    image = (short int*) _mm_malloc(d*N*sizeof(short int),16);
    initimg = (short int*) _mm_malloc(d*N*sizeof(short int),16);
    mass = (double*) _mm_malloc(N*sizeof(double),64);
    sigma = (double*) _mm_malloc(N*sizeof(double),64);
    epsilon = (double*) _mm_malloc(N*sizeof(double),64);
    sqrtepsilon = (double*) _mm_malloc(N*sizeof(double),64);
    Z_array = (int*) _mm_malloc(N*sizeof(int),64);
    
    ifstream resin, rinitin;
    
// If readresflag = false, initialize mass, sigma, and epsilon of all particles,
// give all particles random positions, zero velocities and forces,
// and set Rinit and Rlast to current positions
// This routime is set up for both continuously-polydisperse GLJ/MORSE and bidisperse KOB-ANDERSEN
    srand(time(NULL));
    if (!readresflag) {
        step = 0;
        i = 0;
        if (model == 1 || model == 2) do  {
            trialm = ALU[1] + ((ALU[2]-ALU[1])*rand())/RAND_MAX;
            ranr = (1.0*rand())/RAND_MAX;
            prob = pow(trialm/ALU[1],-d);
            if (ranr < prob) {
                sigma[i] = trialm;
                mass[i] = pow(sigma[i],d);  // Unit-diameter particles have unit mass
                epsilon[i] = pow(sigma[i],d);   // Particles' "charges" are proportional to their volumes
                totpartvol += vs[d]*pow(sigma[i],d);
                i++;
            }
        } while (i < N);
        else if (model == 3) for (i = 0; i < N; i++) {
            mass[i] = 1; // Use to avoid getting "trivially" heterogeneous dynamics
            if (i < Nlarge) {  sigma[i] = sAA;   epsilon[i] = 1;  }
            else {  sigma[i] = 0.88*sAA;   epsilon[i] = 0.5;  }
            totpartvol += vs[d]*pow(sigma[i],d);
        }
// Set cell boundaries
        vol = totpartvol/phi;
        rho = N/vol;
        D = pow(vol,1.0/d);     Dinit = D;
        lo = -D/2;  hi = D/2;
// THEN initialize particle positions
        for (k = 0; k < d*N; k++) {
            RVF[(3*d)*(k/d) + k%d] = lo + (D*rand())/RAND_MAX;
            RVF[(3*d)*(k/d) + k%d + d] = 0;
            RVF[(3*d)*(k/d) + k%d + 2*d] = 0;
            Rlast[k] = RVF[(3*d)*(k/d) + k%d];
            Rinit[k] = RVF[(3*d)*(k/d) + k%d];
            image[k] = 0;   initimg[k] = 0;
        }
    }
// If readresflag = true, read all of the above in from user-specifed restart and Rinit files
    else {
        resin.open(initresfname);
        if (!resin.is_open()) { cout << "Failed to open restart-file " << initresfname << endl;    exit(1);   }
        resin >> step >> lo >> hi;
        D = hi - lo;
        vol = pow(D,d);
        rho = N/vol;
        for (i = 0; i < N; i++) {
            resin >> k >> mass[i] >> sigma[i] >> epsilon[i];
            for (j = 0; j < 2*d; j++) resin >> RVF[3*d*i+j];
            for (j = 0; j < d; j++) resin >> image[d*i+j];
            for (j = 0; j < d; j++) Rlast[d*i+j] = RVF[3*d*i+j];
            totpartvol += vs[d]*pow(sigma[i],d);
        }
        resin.close();
// If not using an Rinit file, set Rinit = R and initimg = image
        if (!strcmp(initRinitfname,"null")) {
            for (k = 0; k < d*N; k++) {
                Rinit[k] = RVF[(3*d)*(k/d) + k%d];
                initimg[k] = image[k];
            }
            Dinit = D;
        }
// If using an Rinit file, get Rinit and initimg from this file
        else if (strcmp(initRinitfname,"null")) {
            rinitin.open(initRinitfname);
            if (!rinitin.is_open()) { cout << "Failed to open Rinit-file " << initRinitfname << endl;    exit(1);   }
            rinitin >> initstep >> initlo >> inithi;
            Dinit = inithi - initlo;
            for (i = 0; i < N; i++) {
                rinitin >> k;
                for (j = 0; j < d; j++) rinitin >> Rinit[d*i+j];
                for (j = 0; j < d; j++) rinitin >> initimg[d*i+j];
            }
            rinitin.close();
        }
    }
    
// Set sqrtepsilon (speeds implementation of the Lorentz-Berthelot mixing rules)
// sqrtepsilon is only used for models1-2
    for (i = 0; i < N; i++) sqrtepsilon[i] = sqrt(epsilon[i]);

// Set maxneigh, initialize the neighbor lists
    rho = N/vol;
    maxneigh = ceil(max(2.5*kissnum[d], 1.5*vs[d]*rho*pow(2*(cut+skin),d)));
    unsigned long long int totneigh = N*maxneigh;
    allneigh = (int*) _mm_malloc(totneigh*sizeof(int),32);
    numneigh = (short int*) _mm_malloc(N*sizeof(short int),16);
    cout << "NL arrays of length " << totneigh << " allocated." << endl;
// If necessary, re-order atom indices from largest to smallest sigma
// if (!readresflag && (model == 1 || model == 2)) reorderatoms();
    if ((model == 1 || model == 2)) reorderatoms();
    return totpartvol;
}


// IDxCompare is used for ordering particle indices in order of decreasing sigma
// See https://stackoverflow.com/questions/25921706/creating-a-vector-of-indices-of-a-sorted-vector
struct IdxCompare {
    const std::vector<double>& target;
    IdxCompare(const std::vector<double>& target): target(target) {}
    bool operator()(int a, int b) const { return target[b] < target[a]; }
};

// Re-order atom indices from lowest to highest particle diameter
// Also need re-order all other quantities like mass, sigma, etc.
void utility::reorderatoms() {
    int i, j;
    double *RVFtmp, *masstmp, *sigmatmp, *epsilontmp, *sqrtepsilontmp;
    short int *imagetmp, *initimgtmp;
    RVFtmp = (double*) _mm_malloc(3*d*N*sizeof(double),64);
    masstmp = (double*) _mm_malloc(N*sizeof(double),64);
    sigmatmp = (double*) _mm_malloc(N*sizeof(double),64);
    epsilontmp = (double*) _mm_malloc(N*sizeof(double),64);
    sqrtepsilontmp = (double*) _mm_malloc(N*sizeof(double),64);
    imagetmp = (short int*) _mm_malloc(d*N*sizeof(short int),16);
    initimgtmp = (short int*) _mm_malloc(d*N*sizeof(short int),16);
    for (i = 0; i < 3*d*N; i++) RVFtmp[i] = RVF[i];
    for (i = 0; i < N; i++) {
        masstmp[i] = mass[i];           sigmatmp[i] = sigma[i];
        epsilontmp[i] = epsilon[i];     sqrtepsilontmp[i] = sqrtepsilon[i];
    }
    for (i = 0; i < d*N; i++) {
        imagetmp[i] = image[i];     initimgtmp[i] = initimg[i];
    }
    vector<double> pdiams;      pdiams.resize(N);
    vector <int> pindices;
    for (i = 0; i < N; i++) {
        pindices.push_back(i);      pdiams[i] = sigma[i];
    }
    sort(pindices.begin(),pindices.end(),IdxCompare(pdiams));
    for (i = 0; i < N; i++) {
        for (j = 0; j < 3*d; j++) RVF[3*d*i + j] = RVFtmp[3*d*pindices[i] + j];
        for (j = 0; j < d; j++) image[d*i + j] = imagetmp[d*pindices[i] + j];
        for (j = 0; j < d; j++) initimg[d*i + j] = initimgtmp[d*pindices[i] + j];
        mass[i] = masstmp[pindices[i]];
        sigma[i] = sigmatmp[pindices[i]];
        epsilon[i] = epsilontmp[pindices[i]];
        sqrtepsilon[i] = sqrtepsilontmp[pindices[i]];
    }
    _mm_free(RVFtmp);       _mm_free(masstmp);      _mm_free(sigmatmp);
    _mm_free(epsilontmp);   _mm_free(sqrtepsilontmp);
    _mm_free(imagetmp);     _mm_free(initimgtmp);
    
    return;
}



// Deallocate all arrays at end of run
void utility::dealloc() {
    _mm_free(RVF);      _mm_free(Rlast);    _mm_free(Rinit);
    _mm_free(image);    _mm_free(initimg);
    _mm_free(allneigh);     _mm_free(numneigh);
    _mm_free(mass);     _mm_free(sigma);    _mm_free(epsilon);
    _mm_free(sqrtepsilon);
    
    return;
}
